from torch.optim.lr_scheduler import MultiStepLR, StepLR, CosineAnnealingLR, CosineAnnealingWarmRestarts, PolynomialLR, \
    LambdaLR, ReduceLROnPlateau

from medlab.registry import LR_SCHEDULERS

LR_SCHEDULERS.register_module(name='MultiStepLR', module=MultiStepLR)
LR_SCHEDULERS.register_module(name='StepLR', module=StepLR)
LR_SCHEDULERS.register_module(name='CosineAnnealingLR', module=CosineAnnealingLR)
LR_SCHEDULERS.register_module(name='CosineAnnealingWarmRestarts', module=CosineAnnealingWarmRestarts)
LR_SCHEDULERS.register_module(name='PolynomialLR', module=PolynomialLR)
LR_SCHEDULERS.register_module(name='LambdaLR', module=LambdaLR)
LR_SCHEDULERS.register_module(name='ReduceLROnPlateau', module=ReduceLROnPlateau)



import os

from monai.data import Dataset
from sklearn.model_selection import train_test_split

from medlab.registry import DATASETS, TRANSFORMS
import configparser
import pandas as pd


@DATASETS.register_module()
class MedAILabClsDataset(Dataset):
    META_INFO = {'classes': {}}

    def __init__(
            self,
            data_root: str,
            subset: str = None,
            val_size: float = 0.2,
            test_size: float = 0,
            seed: int = 1234,
            transforms: dict = None,
            metainfo: str = '.dataset.medailab',
            annotation_file: str = "classification.csv"
    ):
        assert subset in ['train', 'val', 'test', None], "subset类型错误"
        assert val_size >= 0 and val_size < 1
        assert test_size >= 0 and test_size < 1   
        meta_file = os.path.join(data_root, metainfo)
        annotation_file = os.path.join(data_root, annotation_file)
        assert os.path.exists(meta_file), "meta文件不存在"
        assert os.path.exists(annotation_file), "标注文件不存在"

        config = configparser.ConfigParser()
        config.read(meta_file, encoding='utf8')
        assert "classes" in config.sections(), "meta文件错误"
        options = config.options('classes')
        classes = {}
        for opt in options:
            classes[config.get('classes', opt)] = opt
        self.META_INFO['classes'] = classes

        data = pd.read_csv(annotation_file, header=None)
        data.columns = ["images", "labels"]
        data.drop(data[data['labels'] == -1].index, inplace=True)
        images = data["images"]
        labels = data["labels"]

        images = [os.path.join(data_root, "Images", i) for i in images]
    
        assert len(images) > 0, 'No images found in {}'.format(data_root)
        assert len(images) == len(labels), 'The number of images:{} and labels:{} should be equal'.format(len(images),
                                                                                             len(labels))

        if subset is not None and val_size > 0:
            train_x, tmp_x, train_y, tmp_y = train_test_split(images, labels, test_size=val_size+test_size, random_state=seed,
                                                              stratify=labels)
            if subset == 'train':
                images = train_x
                labels = train_y
            else:
                if test_size == 0:
                    images = tmp_x
                    labels = tmp_y
                else:
                    val_x, test_x, val_y, test_y = train_test_split(tmp_x, tmp_y, test_size=test_size/(val_size+test_size), random_state=seed,
                                                              stratify=tmp_y)
                    if subset == 'val':
                        images = val_x
                        labels = val_y
                    else:
                        images = test_x
                        labels = test_y
        data = [{'images': i, 'labels': m} for i, m in zip(images, labels)]
        if transforms is None:
            transforms = []
        transforms = TRANSFORMS.build(transforms)
        super().__init__(data=data, transform=transforms)

    @property
    def classes(self):
        return self.META_INFO['classes']


@DATASETS.register_module()
class MedAILabSegDataset(Dataset):
    META_INFO = {'classes': {}}

    def __init__(
            self,
            data_root: str,
            subset: str = None,
            val_size: float = 0.2,
            test_size: float = 0,
            seed: int = 1234,
            transforms: dict = None,
            metainfo: str = '.dataset.medailab',
            annotation_dir: str = "Segmentation"
    ):
        print("data_root", data_root)
        assert subset in ['train', 'val', 'test', None], "subset类型错误"
        assert val_size >= 0 and val_size < 1
        assert test_size >= 0 and test_size < 1   
        meta_file = os.path.join(data_root, metainfo)
        image_dir = os.path.join(data_root, "Images")
        annotation_dir = os.path.join(data_root, annotation_dir)
        assert os.path.exists(meta_file), "meta文件不存在"
        assert os.path.exists(annotation_dir), "标注目录不存在"

        config = configparser.ConfigParser()
        config.read(meta_file, encoding='utf8')
        assert "classes" in config.sections(), "meta文件错误"
        options = config.options('classes')
        classes = {}
        for opt in options:
            classes[config.get('classes', opt)] = opt
        self.META_INFO['classes'] = classes

        images = []
        labels = []
        for root, dirs, files in os.walk(image_dir):
            for file in files:
                images.append(os.path.join(root, file))
                labels.append(os.path.join(root.replace(image_dir, annotation_dir), file))

        assert len(images) > 0, 'No images found in {}'.format(data_root)
        assert len(images) == len(labels), 'The number of images:{} and labels:{} should be equal'.format(len(images),
                                                                                             len(labels))

        print("数据集长度：", len(images), len(labels))

        if subset is not None and val_size > 0:
            train_x, tmp_x, train_y, tmp_y = train_test_split(images, labels, test_size=val_size+test_size, random_state=seed)
            if subset == 'train':
                images = train_x
                labels = train_y
            else:
                if test_size == 0:
                    images = tmp_x
                    labels = tmp_y
                else:
                    val_x, test_x, val_y, test_y = train_test_split(tmp_x, tmp_y, test_size=test_size/(val_size+test_size), random_state=seed)
                    if subset == 'val':
                        images = val_x
                        labels = val_y
                    else:
                        images = test_x
                        labels = test_y
        data = [{'images': i, 'labels': m} for i, m in zip(images, labels)]
        if transforms is None:
            transforms = []
        transforms = TRANSFORMS.build(transforms)
        super().__init__(data=data, transform=transforms)

    @property
    def classes(self):
        return self.META_INFO['classes']


import glob
import os.path as osp

import pandas as pd
from monai.data import Dataset
from sklearn.model_selection import train_test_split

from medlab.registry import DATASETS, TRANSFORMS


@DATASETS.register_module()
class BaseSegDataset(Dataset):
    def __init__(
            self,
            data_root: str,
            img_dir: str,
            label_dir: str,
            img_suffix: str,
            label_suffix: str,
            subset: str = None,
            test_size: float = 0.2,
            seed: int = 1234,
            transforms: dict = None,
            annotation_file: str = None,
    ):
        """
        base segmentation dataset
        if annotation_file is not None, use annotation file to load data
        else, use subset and test_size to split data

        if subset is None, use all data
        if subset is 'train', use train data
        if subset is 'val', use val data
        :param data_root: dataset root path
        :param img_dir:  image dir
        :param label_dir: label dir
        :param img_suffix: image suffix
        :param label_suffix: label suffix
        :param subset: train or val or None
        :param test_size: test size
        :param seed: random seed
        :param transforms: MONAI data transforms cfg
        :param annotation_file: annotation file path
        """
        assert subset in ['train', 'val', None]

        if annotation_file is not None:
            dataframe = pd.read_csv(osp.join(data_root, annotation_file), header=None, sep=' ')
            images = [osp.join(data_root, i) for i in dataframe[0].to_list()]
            labels = [osp.join(data_root, i) for i in dataframe[1].to_list()]
        else:
            images = sorted(glob.glob(osp.join(data_root, img_dir, '*' + img_suffix)))
            labels = sorted(glob.glob(osp.join(data_root, label_dir, '*' + label_suffix)))

        assert len(images) > 0, 'No images found in {}'.format(osp.join(data_root, img_dir))
        assert len(images) == len(labels), 'The number of images:{} and labels:{} should be equal'.format(len(images),
                                                                                                          len(labels))

        if subset is not None:
            train_x, val_x, train_y, val_y = train_test_split(images, labels, test_size=test_size, random_state=seed)
            if subset == 'train':
                images = train_x
                labels = train_y
            else:
                images = val_x
                labels = val_y

        data = [{'images': i, 'labels': m} for i, m in zip(images, labels)]
        transforms = TRANSFORMS.build(transforms)
        super().__init__(data=data, transform=transforms)



from lightning.pytorch.loggers import TensorBoardLogger, CSVLogger

from medlab.registry import LOGGERS

LOGGERS.register_module(name='TensorBoardLogger', module=TensorBoardLogger)
LOGGERS.register_module(name='CSVLogger', module=CSVLogger)



import argparse
import logging
import os
import os.path as osp
import warnings

from lightning.pytorch import seed_everything
from mmengine.config import Config
from medlab.registry import TRAINERS, TASKS



def parse_args():
    parser = argparse.ArgumentParser(description='Train a segmentor')
    parser.add_argument('config', help='train config file path')
    parser.add_argument('--checkpoint', help='model checkpoint file path', default=None)
    parser.add_argument('--work-dir', help='the dir to save logs and models')
    parser.add_argument(
        '--resume',
        action='store_true',
        default=False,
        help='resume from the latest checkpoint in the work_dir automatically')
    parser.add_argument(
        '--amp',
        action='store_true',
        default=False,
        help='enable automatic-mixed-precision training')

    args = parser.parse_args()

    return args


def train():
    warnings.filterwarnings("ignore")
    logging.disable(logging.WARNING)
    args = parse_args()

    # load config
    cfg = Config.fromfile(args.config)
    print(args.config)
    seed = cfg.get('seed', 1)
    ckpt_path = cfg.get('ckpt_path', None)
    trainer_cfg = cfg.get('trainer', dict())
    seed_everything(seed)

    if args.checkpoint is not None:
        ckpt_path = args.checkpoint

    
    if args.work_dir is not None:
        # update configs according to CLI args if args.work_dir is not None
        trainer_cfg['default_root_dir'] = args.work_dir
    elif trainer_cfg.get('default_root_dir', None) is None:
        # use config filename as default work_dir if cfg.work_dir is None
        trainer_cfg['default_root_dir'] = osp.join('./work_dirs', osp.splitext(osp.basename(args.config))[0])

    work_dir = trainer_cfg['default_root_dir']

    if args.amp is True:
        trainer_cfg.update(dict(precision='16'))

    trainer = TRAINERS.build(trainer_cfg)

    task = TASKS.build(
        dict(
            type=cfg.get('task_type', None),
            model=cfg.get('model', None),
            train_dataloader=cfg.get('train_dataloader', None),
            val_dataloader=cfg.get('val_dataloader', None),
            test_dataloader=None,
            loss_func=cfg.get('loss_func', None),
            optims=cfg.get('optims', None),
            metrics=cfg.get('metrics', None),
            train_cfg=cfg.get('train_cfg', dict()),
            val_cfg=cfg.get('val_cfg', dict()),
            test_cfg=dict(),
            num_classes=cfg.get('num_classes', None)
        )
    )
    os.makedirs(osp.join(work_dir, trainer.time_prefix), exist_ok=True)
    cfg.dump(osp.join(work_dir, trainer.time_prefix, 'config.py'))
    trainer.fit(task, ckpt_path=ckpt_path)


if __name__ == '__main__':
    train()



from typing import Any
from monai.data import decollate_batch
from monai.utils.enums import PostFix

from medlab.registry import TASKS, INFERERS, TRANSFORMS
from .base_task import BaseTask
import torch
import copy
from monai.metrics import ConfusionMatrixMetric
from sklearn.metrics import roc_curve, auc, RocCurveDisplay, confusion_matrix, ConfusionMatrixDisplay
import pathlib
import pandas as pd
import os
import numpy as np
from itertools import cycle
import matplotlib.pyplot as plt
import json
from copy import deepcopy


@TASKS.register_module()
class BaseClsTask(BaseTask):
    def __init__(self, *args, **kwargs):
        """
        base classification task, one input, one output
        """
        super().__init__(*args, **kwargs)
        self.val_inferer = INFERERS.build(self.val_cfg.get('inferer', dict(type='SimpleInferer')))
        self.test_inferer = INFERERS.build(self.test_cfg.get('inferer', dict(type='SimpleInferer')))
        num_classes = kwargs.get('num_classes', None)
        assert num_classes is not None, "num_classes must be specified in model"
        if num_classes == 1:
            self.post_pred_act = TRANSFORMS.build([
                dict(type='ToDevice', device='cpu'),
                dict(type='Activations', sigmoid=True)
            ])
            self.post_pred_cls = TRANSFORMS.build([
                dict(type='AsDiscrete', threshold=0.5)
            ])
            self.post_label = TRANSFORMS.build([dict(type='ToDevice', device='cpu')])

            post_save = []
        else:
            self.post_pred_act = TRANSFORMS.build([
                dict(type='ToDevice', device='cpu'),
                dict(type='Activations', softmax=True)
            ])

            self.post_pred_cls = TRANSFORMS.build([
                dict(type='AsDiscrete', argmax=True, to_onehot=num_classes),
            ])
            self.post_label = TRANSFORMS.build([
                dict(type='AsDiscrete', to_onehot=num_classes),
                dict(type='ToDevice', device='cpu')
            ])
            post_save = [
                dict(type='AsDiscreted', keys='preds', argmax=True)
            ]
        save_dir = self.test_cfg.get('save_dir', None)
        if save_dir is not None:
            post_save.extend([
                dict(
                    type='CopyItemsd',
                    keys=PostFix.meta("images"),
                    times=1,
                    names=PostFix.meta("preds")
                ),
                dict(
                    type='SaveClassificationd',
                    keys='preds',
                    saver=None,
                    meta_keys=None,
                    output_dir=save_dir,
                    filename='predictions.csv',
                    delimiter=",",
                    overwrite=True)
            ])
        self.post_save = TRANSFORMS.build(post_save)

        self.train_metrics_key = copy.deepcopy(self.metrics_key)
        self.train_metrics = copy.deepcopy(self.metrics)

        self.val_metrics_key = copy.deepcopy(self.metrics_key)
        self.val_metrics = copy.deepcopy(self.metrics)

        if 'ROCAUCMetric' in self.metrics_key:
            index_train = self.train_metrics_key.index('ROCAUCMetric')
            index_val = self.train_metrics_key.index('ROCAUCMetric')
            self.train_metrics_key.pop(index_train)
            self.val_metrics_key.pop(index_val)
            self.train_roc_auc_metric = self.train_metrics.pop(index_train)
            self.val_roc_auc_metric = self.val_metrics.pop(index_val)
        else:
            self.train_roc_auc_metric = None
            self.val_roc_auc_metric = None

        self.training_step_loss = []
        self.validation_step_loss = []
        self.test_names = []
        self.test_labels = []
        self.test_preds = []

    def forward(self, x):
        return self._model(x)

    def training_step(self, batch, batch_idx):
        """
        training step for classification task
        :param batch: batch data
        :param batch_idx: batch index
        :return: loss
        """
        images = batch['images']
        labels = batch['labels']

        batch_size = images.shape[0]
        outputs = self.forward(images)


        loss = self.loss_func(outputs, labels)

        if isinstance(outputs, (tuple, list)):
            outputs = outputs[0]

        outputs = [self.post_pred_act(i) for i in decollate_batch(outputs)]
        labels = [self.post_label(i) for i in decollate_batch(labels)]

        if self.train_roc_auc_metric is not None:
            self.train_roc_auc_metric(outputs, labels)
        
        outputs = [self.post_pred_cls(i) for i in outputs]

        if len(self.train_metrics) > 0:
            for metric in self.train_metrics:
                metric(outputs, labels)

        self.training_step_loss.append(loss)

        # self.log("train_step_loss", loss.item(), sync_dist=True, batch_size=batch_size)
        return loss

    def on_train_epoch_end(self):
        self.log_dict(self.parse_train_metrics(), sync_dist=True)
        epoch_mean = torch.stack(self.training_step_loss).mean()
        self.log("train_loss", epoch_mean.item(), sync_dist=True)
        self.training_step_loss.clear()

    def validation_step(self, batch, batch_idx):
        """
        validation step for classification task
        :param batch: batch data
        :param batch_idx: batch index
        :return: None
        """
        images = batch['images']
        labels = batch['labels']

        batch_size = images.shape[0]
        outputs = self.val_inferer(inputs=images, network=self.forward)

        loss = self.loss_func(outputs, labels)


        if isinstance(outputs, (tuple, list)):
            outputs = outputs[0]

        outputs = [self.post_pred_act(i) for i in decollate_batch(outputs)]
        labels = [self.post_label(i) for i in decollate_batch(labels, detach=False)]

        if self.val_roc_auc_metric is not None:
            self.val_roc_auc_metric(outputs, labels)

        outputs = [self.post_pred_cls(i) for i in outputs]

        if len(self.val_metrics) > 0:
            for metric in self.val_metrics:
                metric(outputs, labels)
        
        self.validation_step_loss.append(loss)

        # self.log('val_step_loss', loss.item(), sync_dist=True, batch_size=batch_size)

    def on_validation_epoch_end(self):
        """
        validation epoch end hook, parse and log metrics
        """
        self.log_dict(self.parse_val_metrics(), sync_dist=True)
        epoch_mean = torch.stack(self.validation_step_loss).mean()
        self.log("val_loss", epoch_mean.item(), sync_dist=True)
        self.validation_step_loss.clear()


    def test_step(self, batch, batch_idx):
        """
        test step for classification task, save predictions to csv file
        :param batch: batch data
        :param batch_idx: batch index
        :return:
        """
 
        images = batch["images"]
        labels = batch["labels"]

        outputs = self.test_inferer(inputs=images, network=self.forward)

        if isinstance(outputs, (tuple, list)):
            outputs = outputs[0]

        batch["preds"] = outputs

        outputs = [self.post_pred_act(i) for i in decollate_batch(outputs)]
        labels = [self.post_label(i) for i in decollate_batch(labels)]

        if self.val_roc_auc_metric is not None:
            self.val_roc_auc_metric(outputs, labels)

        outputs = [self.post_pred_cls(i) for i in outputs]

        if len(self.val_metrics) > 0:
            for metric in self.val_metrics:
                metric(outputs, labels)

        for i in decollate_batch(batch):
            # self.post_save(i)
            filename = i['images_meta_dict']['filename_or_obj']
            # filename = pathlib.Path(filename).name
            self.test_names.append(filename)
            self.test_labels.append(i['labels'])
            self.test_preds.append(torch.argmax(i['preds']).item())

        self.log("test_step", batch_idx, sync_dist=True, batch_size=images.shape[0])

    def on_test_epoch_end(self):
        """
        test epoch end hook, parse metrics
        """
    
        df = pd.DataFrame({'name': self.test_names, 'label': self.test_labels, 'pred': self.test_preds})
        df.to_csv(os.path.join(self.test_cfg.get('save_dir'), "test.csv"), index=False)

        if self.val_roc_auc_metric is not None:
            y_pred_prob, y = self.val_roc_auc_metric.get_buffer()
            y_pred = torch.argmax(y_pred_prob, dim=1, keepdim=True)
            y_pred_ont_hot = torch.zeros_like(y_pred_prob)
            y_pred_ont_hot.scatter_(1, y_pred, 1)

            meta = self._test_dataloader.dataset.META_INFO['classes']
            target_names = [meta[i] for i in sorted(meta.keys())]
            num_classes = len(target_names)

            # 绘制混淆矩阵
            # 分别设置图片大小和画布大小，保证图片中的内容不被截断
            fig, ax = plt.subplots(figsize=(8, 6))
            cmap = plt.cm.Blues
            cm = confusion_matrix(torch.argmax(y, dim=1).cpu().numpy().ravel(), y_pred.cpu().numpy().ravel())
            disp = ConfusionMatrixDisplay(confusion_matrix=cm, display_labels=target_names)
            disp.plot(include_values=True, cmap=cmap, ax=ax, xticks_rotation='horizontal')
            ax.set_title("Confusion Matrix")
            plt.savefig(os.path.join(self.test_cfg.get('save_dir'), "confusion_matrix.png"))


            cfm = ConfusionMatrixMetric(
                metric_name=("accuracy", "sensitivity", "specificity", "precision", "recall", "f1_score")
            )
            cfm(y_pred_ont_hot, y)

            acc, sen, spe, pre, rec, f1 = [i.item() for i in cfm.aggregate()]
            test_rst = {}
            test_rst['accuracy'] = acc
            test_rst['sensitivity'] = sen
            test_rst['specificity'] = spe
            test_rst['precision'] = pre
            test_rst['recall'] = rec
            test_rst['f1_score'] = f1

            df[df["pred"] != df["label"]].to_csv(os.path.join(self.test_cfg.get('save_dir'), "error.csv"), index=False)
            json.dump(test_rst, open(os.path.join(self.test_cfg.get('save_dir'), "metric.json"), 'w'), indent=4)
            
            
            y_pred_prob = y_pred_prob.cpu().numpy()
            y = y.cpu().numpy()

            if num_classes == 2:
                y = y[:, 1].ravel()
                y_pred_prob = y_pred_prob[:, 1].ravel()
                fpr, tpr, thresholds = roc_curve(y, y_pred_prob)
                roc_auc = auc(fpr, tpr)
                fig, ax = plt.subplots(figsize=(6, 6))
                RocCurveDisplay.from_predictions(
                    y,
                    y_pred_prob,
                    name=f"ROC curve for {target_names[1]}",
                    color="deeppink",
                    ax=ax
                )
                plt.plot([0, 1], [0, 1], color="navy", linestyle="--")
                _ = ax.set(
                    xlabel="False Positive Rate",
                    ylabel="True Positive Rate",
                    title="Receiver Operating Characteristic",
                )
                ax.set_xlim([-0.01, 1.01])
                ax.set_ylim([-0.01, 1.01])
                fig.savefig(os.path.join(self.test_cfg.get('save_dir'), "roc.png"))
            else:
                fpr, tpr, roc_auc = dict(), dict(), dict()
                fpr["micro"], tpr["micro"], _ = roc_curve(y.ravel(), y_pred_prob.ravel())
                roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

                for i in range(num_classes):
                    fpr[i], tpr[i], _ = roc_curve(y[:, i], y_pred_prob[:, i])
                    roc_auc[i] = auc(fpr[i], tpr[i])
                fpr_grid = np.linspace(0.0, 1.0, 1000)
                mean_tpr = np.zeros_like(fpr_grid)
                for i in range(num_classes):
                    mean_tpr += np.interp(fpr_grid, fpr[i], tpr[i])  # linear interpolation
                mean_tpr /= num_classes

                fpr["macro"] = fpr_grid
                tpr["macro"] = mean_tpr
                roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])
                fig, ax = plt.subplots(figsize=(6, 6))

                plt.plot(
                    fpr["micro"],
                    tpr["micro"],
                    label=f"micro-average ROC curve (AUC = {roc_auc['micro']:.2f})",
                    color="deeppink",
                    linestyle=":",
                    linewidth=4,
                )

                plt.plot(
                    fpr["macro"],
                    tpr["macro"],
                    label=f"macro-average ROC curve (AUC = {roc_auc['macro']:.2f})",
                    color="navy",
                    linestyle=":",
                    linewidth=4,
                )

                colors = cycle(["aqua", "darkorange", "cornflowerblue"])
                
                for class_id, color in zip(range(num_classes), colors):
                    RocCurveDisplay.from_predictions(
                        y[:, class_id],
                        y_pred_prob[:, class_id],
                        name=f"ROC curve for {target_names[class_id]}",
                        color=color,
                        ax=ax
                        # plot_chance_level=(class_id == 2),
                    )

                plt.plot([0, 1], [0, 1], color="navy", linestyle="--")
                
                _ = ax.set(
                xlabel="False Positive Rate",
                ylabel="True Positive Rate",
                title="Extension of Receiver Operating Characteristic\nto One-vs-Rest multiclass",
                )
                

                ax.set_xlim([-0.01, 1.01])
                ax.set_ylim([-0.01, 1.01])

                fig.savefig(os.path.join(self.test_cfg.get('save_dir'), "roc.png"))
       
       
        

    def predict_step(self, batch, batch_idx, dataloader_idx=0) -> Any:
        images = batch["images"]
        outputs = self.test_inferer(inputs=images, network=self.forward)
        if isinstance(outputs, (tuple, list)):
            outputs = outputs[0]
        batch["preds"] = outputs
        for i in decollate_batch(batch):
            self.test_names.append(i['images_meta_dict']['filename_or_obj'])
            self.test_preds.append(torch.argmax(i['preds']).item())


    def on_predict_epoch_end(self):
        df = pd.DataFrame({'name': self.test_names, 'pred': self.test_preds})
        df.to_csv(os.path.join(self.test_cfg.get('save_dir'), "predictions.csv"), index=False, encoding='utf-8')

    def parse_train_metrics(self):
        """
        parse metrics to dict
        :return: metrics dict
        """
        value_dict = {}
        values = []
        if self.train_roc_auc_metric is not None:
            value_dict['ROCAUCMetric'] = self.train_roc_auc_metric.aggregate()
            self.train_roc_auc_metric.reset()

        for metric in self.train_metrics:
            value = metric.aggregate()

            if isinstance(value, list):
                values.extend([v.item() for v in value])
            else:
                values.append(value.item())
        value_dict.update(dict(zip(self.train_metrics_key, values)))

        for metric in self.metrics:
            metric.reset()
        
        value_dict = {f'train_{k}': v for k, v in value_dict.items()}

        return value_dict

    def parse_val_metrics(self):
        """
        parse metrics to dict
        :return: metrics dict
        """
        value_dict = {}
        values = []
        if self.val_roc_auc_metric is not None:
            value_dict['ROCAUCMetric'] = self.val_roc_auc_metric.aggregate()
            self.val_roc_auc_metric.reset()

        for metric in self.val_metrics:
            value = metric.aggregate()

            if isinstance(value, list):
                values.extend([v.item() for v in value])
            else:
                values.append(value.item())
        value_dict.update(dict(zip(self.val_metrics_key, values)))

        for metric in self.metrics:
            metric.reset()
        
        value_dict = {f'val_{k}': v for k, v in value_dict.items()}

        return value_dict


from monai.inferers import SimpleInferer, SlidingWindowInferer

from medlab.registry import INFERERS

INFERERS.register_module(name='SimpleInferer', module=SimpleInferer)
INFERERS.register_module(name='SlidingWindowInferer', module=SlidingWindowInferer)



from torch.optim import SGD, Adam, AdamW, RMSprop

from medlab.registry import OPTIMIZERS

OPTIMIZERS.register_module(name='SGD', module=SGD)
OPTIMIZERS.register_module(name='Adam', module=Adam)
OPTIMIZERS.register_module(name='AdamW', module=AdamW)
OPTIMIZERS.register_module(name='RMSprop', module=RMSprop)



import argparse
import logging
import os.path as osp
import warnings

from lightning.pytorch import seed_everything
from mmengine.config import Config

from medlab.registry import TRAINERS, TASKS


def parse_args():
    parser = argparse.ArgumentParser(description='Test a model')
    parser.add_argument('checkpoint', help='model checkpoint file path')
    parser.add_argument('--config', help='train config file path')
    parser.add_argument('--save-dir', help='the dir to save logs and models')

    args = parser.parse_args()

    return args


def test():
    warnings.filterwarnings("ignore")
    logging.disable(logging.WARNING)
    args = parse_args()

    # load config
    if args.config is not None:
        cfg_path = args.config
    else:
        cfg_path = osp.join(osp.dirname(osp.dirname(args.checkpoint)), 'config.py')
        
    assert osp.exists(cfg_path)

    cfg = Config.fromfile(cfg_path)

    seed = cfg.get('seed', 1)
    ckpt_path = cfg.get('ckpt_path', None)
    trainer_cfg = cfg.get('trainer', dict())
    seed_everything(seed)

    if args.checkpoint is not None:
        ckpt_path = args.checkpoint

    if args.save_dir is not None:
        # update configs according to CLI args if args.work_dir is not None
        save_dir = args.save_dir
    else:
        save_dir = cfg.get('save_dir', osp.join(osp.dirname(ckpt_path), osp.splitext(osp.basename(ckpt_path))[0]))

    trainer_cfg['default_root_dir'] = save_dir

    trainer = TRAINERS.build(trainer_cfg)

    test_cfg = cfg.get('test_cfg', dict())
    test_cfg.update(dict(save_dir=save_dir))

    
    task = TASKS.build(
        dict(
            type=cfg.get('task_type', None),
            model=cfg.get('model', None),
            train_dataloader=None,
            val_dataloader=None,
            test_dataloader=cfg.get('test_dataloader', None),
            loss_func=cfg.get('loss_func', None),
            optims=None,
            metrics=cfg.get('metrics', None),
            train_cfg=dict(),
            val_cfg=dict(),
            test_cfg=test_cfg,
            num_classes=cfg.get('num_classes', None)
        )
    )
    trainer.test(task, ckpt_path=ckpt_path)

if __name__ == '__main__':
    test()



from monai.transforms import *
from medlab.registry import TRANSFORMS

# https://docs.monai.io/en/stable/transforms.html#module-monai.transforms

# Generic Interfaces
# TRANSFORMS.register_module(name='Transform', module=Transform)
TRANSFORMS.register_module(name='MapTransform', module=MapTransform)
TRANSFORMS.register_module(name='RandomizableTrait', module=RandomizableTrait)
TRANSFORMS.register_module(name='LazyTrait', module=LazyTrait)
TRANSFORMS.register_module(name='MultiSampleTrait', module=MultiSampleTrait)
TRANSFORMS.register_module(name='LazyTransform', module=LazyTransform)
TRANSFORMS.register_module(name='Compose', module=Compose)
TRANSFORMS.register_module(name='InvertibleTransform', module=InvertibleTransform)
TRANSFORMS.register_module(name='TraceableTransform', module=TraceableTransform)
TRANSFORMS.register_module(name='BatchInverseTransform', module=BatchInverseTransform)
TRANSFORMS.register_module(name='Decollated', module=Decollated)
TRANSFORMS.register_module(name='OneOf', module=OneOf)
TRANSFORMS.register_module(name='RandomOrder', module=RandomOrder)

# Vanilla Transforms
# Crop and Pad
TRANSFORMS.register_module(name='PadListDataCollate', module=PadListDataCollate)
TRANSFORMS.register_module(name='Pad', module=Pad)
TRANSFORMS.register_module(name='BorderPad', module=BorderPad)
TRANSFORMS.register_module(name='DivisiblePad', module=DivisiblePad)
TRANSFORMS.register_module(name='Crop', module=Crop)
TRANSFORMS.register_module(name='SpatialCrop', module=SpatialCrop)
TRANSFORMS.register_module(name='CenterSpatialCrop', module=CenterSpatialCrop)
TRANSFORMS.register_module(name='RandSpatialCrop', module=RandSpatialCrop)
TRANSFORMS.register_module(name='RandSpatialCropSamples', module=RandSpatialCropSamples)
TRANSFORMS.register_module(name='CropForeground', module=CropForeground)
TRANSFORMS.register_module(name='RandWeightedCrop', module=RandWeightedCrop)
TRANSFORMS.register_module(name='RandCropByPosNegLabel', module=RandCropByPosNegLabel)
TRANSFORMS.register_module(name='RandCropByLabelClasses', module=RandCropByLabelClasses)
TRANSFORMS.register_module(name='ResizeWithPadOrCrop', module=ResizeWithPadOrCrop)
TRANSFORMS.register_module(name='BoundingRect', module=BoundingRect)
TRANSFORMS.register_module(name='RandScaleCrop', module=RandScaleCrop)
TRANSFORMS.register_module(name='CenterScaleCrop', module=CenterScaleCrop)

# Intensity
TRANSFORMS.register_module(name='RandGaussianNoise', module=RandGaussianNoise)
TRANSFORMS.register_module(name='ShiftIntensity', module=ShiftIntensity)
TRANSFORMS.register_module(name='RandShiftIntensity', module=RandShiftIntensity)
TRANSFORMS.register_module(name='StdShiftIntensity', module=StdShiftIntensity)
TRANSFORMS.register_module(name='RandStdShiftIntensity', module=RandStdShiftIntensity)
TRANSFORMS.register_module(name='ScaleIntensity', module=ScaleIntensity)
TRANSFORMS.register_module(name='RandScaleIntensity', module=RandScaleIntensity)
TRANSFORMS.register_module(name='NormalizeIntensity', module=NormalizeIntensity)
TRANSFORMS.register_module(name='ThresholdIntensity', module=ThresholdIntensity)
TRANSFORMS.register_module(name='ScaleIntensityRange', module=ScaleIntensityRange)
TRANSFORMS.register_module(name='ScaleIntensityRangePercentiles', module=ScaleIntensityRangePercentiles)
TRANSFORMS.register_module(name='AdjustContrast', module=AdjustContrast)
TRANSFORMS.register_module(name='RandAdjustContrast', module=RandAdjustContrast)
TRANSFORMS.register_module(name='MaskIntensity', module=MaskIntensity)
TRANSFORMS.register_module(name='SavitzkyGolaySmooth', module=SavitzkyGolaySmooth)
TRANSFORMS.register_module(name='MedianSmooth', module=MedianSmooth)
TRANSFORMS.register_module(name='GaussianSmooth', module=GaussianSmooth)
TRANSFORMS.register_module(name='RandGaussianSmooth', module=RandGaussianSmooth)
TRANSFORMS.register_module(name='GaussianSharpen', module=GaussianSharpen)
TRANSFORMS.register_module(name='RandGaussianSharpen', module=RandGaussianSharpen)
TRANSFORMS.register_module(name='RandHistogramShift', module=RandHistogramShift)
TRANSFORMS.register_module(name='DetectEnvelope', module=DetectEnvelope)
TRANSFORMS.register_module(name='GibbsNoise', module=GibbsNoise)
TRANSFORMS.register_module(name='RandGibbsNoise', module=RandGibbsNoise)
TRANSFORMS.register_module(name='KSpaceSpikeNoise', module=KSpaceSpikeNoise)
TRANSFORMS.register_module(name='RandKSpaceSpikeNoise', module=RandKSpaceSpikeNoise)
TRANSFORMS.register_module(name='RandCoarseTransform', module=RandCoarseTransform)
TRANSFORMS.register_module(name='RandCoarseDropout', module=RandCoarseDropout)
TRANSFORMS.register_module(name='RandCoarseShuffle', module=RandCoarseShuffle)
TRANSFORMS.register_module(name='HistogramNormalize', module=HistogramNormalize)
TRANSFORMS.register_module(name='ComputeHoVerMaps', module=ComputeHoVerMaps)

# IO
TRANSFORMS.register_module(name='LoadImage', module=LoadImage)
TRANSFORMS.register_module(name='SaveImage', module=SaveImage)

# NVIDIA Tool Extension (NVTX)
TRANSFORMS.register_module(name='RangePush', module=RangePush)
TRANSFORMS.register_module(name='RandRangePush', module=RandRangePush)
TRANSFORMS.register_module(name='RangePop', module=RangePop)
TRANSFORMS.register_module(name='RandRangePop', module=RandRangePop)
TRANSFORMS.register_module(name='Mark', module=Mark)
TRANSFORMS.register_module(name='RandMark', module=RandMark)

# Post-processing
TRANSFORMS.register_module(name='Activations', module=Activations)
TRANSFORMS.register_module(name='AsDiscrete', module=AsDiscrete)
TRANSFORMS.register_module(name='KeepLargestConnectedComponent', module=KeepLargestConnectedComponent)
TRANSFORMS.register_module(name='RemoveSmallObjects', module=RemoveSmallObjects)
TRANSFORMS.register_module(name='LabelFilter', module=LabelFilter)
TRANSFORMS.register_module(name='FillHoles', module=FillHoles)
TRANSFORMS.register_module(name='LabelToContour', module=LabelToContour)
TRANSFORMS.register_module(name='MeanEnsemble', module=MeanEnsemble)
TRANSFORMS.register_module(name='ProbNMS', module=ProbNMS)
TRANSFORMS.register_module(name='SobelGradients', module=SobelGradients)
TRANSFORMS.register_module(name='VoteEnsemble', module=VoteEnsemble)

# Signal
TRANSFORMS.register_module(name='SignalRandDrop', module=SignalRandDrop)
TRANSFORMS.register_module(name='SignalRandScale', module=SignalRandScale)
TRANSFORMS.register_module(name='SignalRandShift', module=SignalRandShift)
TRANSFORMS.register_module(name='SignalRandAddSine', module=SignalRandAddSine)
TRANSFORMS.register_module(name='SignalRandAddSquarePulse', module=SignalRandAddSquarePulse)
TRANSFORMS.register_module(name='SignalRandAddGaussianNoise', module=SignalRandAddGaussianNoise)
TRANSFORMS.register_module(name='SignalRandAddSinePartial', module=SignalRandAddSinePartial)
TRANSFORMS.register_module(name='SignalRandAddSquarePulsePartial', module=SignalRandAddSquarePulsePartial)
TRANSFORMS.register_module(name='SignalFillEmpty', module=SignalFillEmpty)
TRANSFORMS.register_module(name='SignalRemoveFrequency', module=SignalRemoveFrequency)
TRANSFORMS.register_module(name='SignalContinuousWavelet', module=SignalContinuousWavelet)

# Spatial
TRANSFORMS.register_module(name='SpatialResample', module=SpatialResample)
TRANSFORMS.register_module(name='ResampleToMatch', module=ResampleToMatch)
TRANSFORMS.register_module(name='Spacing', module=Spacing)
TRANSFORMS.register_module(name='Orientation', module=Orientation)
TRANSFORMS.register_module(name='RandRotate', module=RandRotate)
TRANSFORMS.register_module(name='RandFlip', module=RandFlip)
TRANSFORMS.register_module(name='RandAxisFlip', module=RandAxisFlip)
TRANSFORMS.register_module(name='RandZoom', module=RandZoom)
TRANSFORMS.register_module(name='Affine', module=Affine)
TRANSFORMS.register_module(name='Resample', module=Resample)
TRANSFORMS.register_module(name='RandAffine', module=RandAffine)
TRANSFORMS.register_module(name='RandDeformGrid', module=RandDeformGrid)
TRANSFORMS.register_module(name='AffineGrid', module=AffineGrid)
TRANSFORMS.register_module(name='RandAffineGrid', module=RandAffineGrid)
TRANSFORMS.register_module(name='GridDistortion', module=GridDistortion)
TRANSFORMS.register_module(name='RandGridDistortion', module=RandGridDistortion)
TRANSFORMS.register_module(name='Rand2DElastic', module=Rand2DElastic)
TRANSFORMS.register_module(name='Rand3DElastic', module=Rand3DElastic)
TRANSFORMS.register_module(name='Rotate90', module=Rotate90)
TRANSFORMS.register_module(name='RandRotate90', module=RandRotate90)
TRANSFORMS.register_module(name='Flip', module=Flip)
TRANSFORMS.register_module(name='Resize', module=Resize)
TRANSFORMS.register_module(name='Rotate', module=Rotate)
TRANSFORMS.register_module(name='Zoom', module=Zoom)
TRANSFORMS.register_module(name='GridPatch', module=GridPatch)
TRANSFORMS.register_module(name='RandGridPatch', module=RandGridPatch)
TRANSFORMS.register_module(name='GridSplit', module=GridSplit)

# Smooth Field
TRANSFORMS.register_module(name='RandSmoothFieldAdjustContrast', module=RandSmoothFieldAdjustContrast)
TRANSFORMS.register_module(name='RandSmoothFieldAdjustIntensity', module=RandSmoothFieldAdjustIntensity)
TRANSFORMS.register_module(name='RandSmoothDeform', module=RandSmoothDeform)

# Utility
TRANSFORMS.register_module(name='Identity', module=Identity)
# TRANSFORMS.register_module(name='AsChannelFirst', module=AsChannelFirst)
TRANSFORMS.register_module(name='AsChannelLast', module=AsChannelLast)
# TRANSFORMS.register_module(name='AddChannel', module=AddChannel)
TRANSFORMS.register_module(name='EnsureChannelFirst', module=EnsureChannelFirst)
TRANSFORMS.register_module(name='RepeatChannel', module=RepeatChannel)
TRANSFORMS.register_module(name='SplitDim', module=SplitDim)
# TRANSFORMS.register_module(name='SplitChannel', module=SplitChannel)
TRANSFORMS.register_module(name='CastToType', module=CastToType)
TRANSFORMS.register_module(name='ToTensor', module=ToTensor)
TRANSFORMS.register_module(name='ToNumpy', module=ToNumpy)
TRANSFORMS.register_module(name='ToCupy', module=ToCupy)
TRANSFORMS.register_module(name='Transpose', module=Transpose)
TRANSFORMS.register_module(name='SqueezeDim', module=SqueezeDim)
TRANSFORMS.register_module(name='DataStats', module=DataStats)
TRANSFORMS.register_module(name='SimulateDelay', module=SimulateDelay)
TRANSFORMS.register_module(name='Lambda', module=Lambda)
TRANSFORMS.register_module(name='RandLambda', module=RandLambda)
TRANSFORMS.register_module(name='RemoveRepeatedChannel', module=RemoveRepeatedChannel)
TRANSFORMS.register_module(name='LabelToMask', module=LabelToMask)
TRANSFORMS.register_module(name='FgBgToIndices', module=FgBgToIndices)
TRANSFORMS.register_module(name='ClassesToIndices', module=ClassesToIndices)
TRANSFORMS.register_module(name='AddExtremePointsChannel', module=AddExtremePointsChannel)
TRANSFORMS.register_module(name='TorchVision', module=TorchVision)
TRANSFORMS.register_module(name='MapLabelValue', module=MapLabelValue)
TRANSFORMS.register_module(name='EnsureType', module=EnsureType)
TRANSFORMS.register_module(name='IntensityStats', module=IntensityStats)
TRANSFORMS.register_module(name='ToDevice', module=ToDevice)
TRANSFORMS.register_module(name='CuCIM', module=CuCIM)
TRANSFORMS.register_module(name='RandCuCIM', module=RandCuCIM)
TRANSFORMS.register_module(name='AddCoordinateChannels', module=AddCoordinateChannels)

# Dictionary Transforms

# Crop and Pad (Dict)
TRANSFORMS.register_module(name='Padd', module=Padd)
TRANSFORMS.register_module(name='SpatialPadd', module=SpatialPadd)
TRANSFORMS.register_module(name='BorderPadd', module=BorderPadd)
TRANSFORMS.register_module(name='DivisiblePadd', module=DivisiblePadd)
TRANSFORMS.register_module(name='Cropd', module=Cropd)
TRANSFORMS.register_module(name='RandCropd', module=RandCropd)
TRANSFORMS.register_module(name='SpatialCropd', module=SpatialCropd)
TRANSFORMS.register_module(name='CenterSpatialCropd', module=CenterSpatialCropd)
TRANSFORMS.register_module(name='RandSpatialCropd', module=RandSpatialCropd)
TRANSFORMS.register_module(name='RandSpatialCropSamplesd', module=RandSpatialCropSamplesd)
TRANSFORMS.register_module(name='CropForegroundd', module=CropForegroundd)
TRANSFORMS.register_module(name='RandWeightedCropd', module=RandWeightedCropd)
TRANSFORMS.register_module(name='RandCropByPosNegLabeld', module=RandCropByPosNegLabeld)
TRANSFORMS.register_module(name='RandCropByLabelClassesd', module=RandCropByLabelClassesd)
TRANSFORMS.register_module(name='ResizeWithPadOrCropd', module=ResizeWithPadOrCropd)
TRANSFORMS.register_module(name='BoundingRectd', module=BoundingRectd)
TRANSFORMS.register_module(name='RandScaleCropd', module=RandScaleCropd)
TRANSFORMS.register_module(name='CenterScaleCropd', module=CenterScaleCropd)

# Intensity (Dict)
TRANSFORMS.register_module(name='RandGaussianNoised', module=RandGaussianNoised)
TRANSFORMS.register_module(name='ShiftIntensityd', module=ShiftIntensityd)
TRANSFORMS.register_module(name='RandShiftIntensityd', module=RandShiftIntensityd)
TRANSFORMS.register_module(name='StdShiftIntensityd', module=StdShiftIntensityd)
TRANSFORMS.register_module(name='RandStdShiftIntensityd', module=RandStdShiftIntensityd)
TRANSFORMS.register_module(name='RandBiasFieldd', module=RandBiasFieldd)
TRANSFORMS.register_module(name='ScaleIntensityd', module=ScaleIntensityd)
TRANSFORMS.register_module(name='RandScaleIntensityd', module=RandScaleIntensityd)
TRANSFORMS.register_module(name='NormalizeIntensityd', module=NormalizeIntensityd)
TRANSFORMS.register_module(name='ThresholdIntensityd', module=ThresholdIntensityd)
TRANSFORMS.register_module(name='ScaleIntensityRanged', module=ScaleIntensityRanged)
TRANSFORMS.register_module(name='GibbsNoised', module=GibbsNoised)
TRANSFORMS.register_module(name='RandGibbsNoised', module=RandGibbsNoised)
TRANSFORMS.register_module(name='KSpaceSpikeNoised', module=KSpaceSpikeNoised)
TRANSFORMS.register_module(name='RandKSpaceSpikeNoised', module=RandKSpaceSpikeNoised)
TRANSFORMS.register_module(name='RandRicianNoised', module=RandRicianNoised)
TRANSFORMS.register_module(name='ScaleIntensityRangePercentilesd', module=ScaleIntensityRangePercentilesd)
TRANSFORMS.register_module(name='AdjustContrastd', module=AdjustContrastd)
TRANSFORMS.register_module(name='RandAdjustContrastd', module=RandAdjustContrastd)
TRANSFORMS.register_module(name='MaskIntensityd', module=MaskIntensityd)
TRANSFORMS.register_module(name='SavitzkyGolaySmoothd', module=SavitzkyGolaySmoothd)
TRANSFORMS.register_module(name='MedianSmoothd', module=MedianSmoothd)
TRANSFORMS.register_module(name='GaussianSmoothd', module=GaussianSmoothd)
TRANSFORMS.register_module(name='RandGaussianSmoothd', module=RandGaussianSmoothd)
TRANSFORMS.register_module(name='GaussianSharpend', module=GaussianSharpend)
TRANSFORMS.register_module(name='RandGaussianSharpend', module=RandGaussianSharpend)
TRANSFORMS.register_module(name='RandHistogramShiftd', module=RandHistogramShiftd)
TRANSFORMS.register_module(name='RandCoarseDropoutd', module=RandCoarseDropoutd)
TRANSFORMS.register_module(name='RandCoarseShuffled', module=RandCoarseShuffled)
TRANSFORMS.register_module(name='HistogramNormalized', module=HistogramNormalized)
TRANSFORMS.register_module(name='ForegroundMaskd', module=ForegroundMaskd)
TRANSFORMS.register_module(name='ComputeHoVerMapsd', module=ComputeHoVerMapsd)

# IO (Dict)
TRANSFORMS.register_module(name='LoadImaged', module=LoadImaged)
TRANSFORMS.register_module(name='SaveImaged', module=SaveImaged)

# Post-processing (Dict)
TRANSFORMS.register_module(name='Activationsd', module=Activationsd)
TRANSFORMS.register_module(name='AsDiscreted', module=AsDiscreted)
TRANSFORMS.register_module(name='KeepLargestConnectedComponentd', module=KeepLargestConnectedComponentd)
TRANSFORMS.register_module(name='RemoveSmallObjectsd', module=RemoveSmallObjectsd)
TRANSFORMS.register_module(name='LabelFilterd', module=LabelFilterd)
TRANSFORMS.register_module(name='FillHolesd', module=FillHolesd)
TRANSFORMS.register_module(name='LabelToContourd', module=LabelToContourd)
TRANSFORMS.register_module(name='Ensembled', module=Ensembled)
TRANSFORMS.register_module(name='MeanEnsembled', module=MeanEnsembled)
TRANSFORMS.register_module(name='VoteEnsembled', module=VoteEnsembled)
TRANSFORMS.register_module(name='Invertd', module=Invertd)
TRANSFORMS.register_module(name='SaveClassificationd', module=SaveClassificationd)
TRANSFORMS.register_module(name='ProbNMSd', module=ProbNMSd)
TRANSFORMS.register_module(name='SobelGradientsd', module=SobelGradientsd)

# Spatial (Dict)
TRANSFORMS.register_module(name='SpatialResampled', module=SpatialResampled)
TRANSFORMS.register_module(name='ResampleToMatchd', module=ResampleToMatchd)
TRANSFORMS.register_module(name='Spacingd', module=Spacingd)
TRANSFORMS.register_module(name='Orientationd', module=Orientationd)
TRANSFORMS.register_module(name='Flipd', module=Flipd)
TRANSFORMS.register_module(name='RandFlipd', module=RandFlipd)
TRANSFORMS.register_module(name='RandAxisFlipd', module=RandAxisFlipd)
TRANSFORMS.register_module(name='Rotated', module=Rotated)
TRANSFORMS.register_module(name='RandRotated', module=RandRotated)
TRANSFORMS.register_module(name='Zoomd', module=Zoomd)
TRANSFORMS.register_module(name='RandZoomd', module=RandZoomd)
TRANSFORMS.register_module(name='GridPatchd', module=GridPatchd)
TRANSFORMS.register_module(name='RandGridPatchd', module=RandGridPatchd)
TRANSFORMS.register_module(name='GridSplitd', module=GridSplitd)
TRANSFORMS.register_module(name='RandRotate90d', module=RandRotate90d)
TRANSFORMS.register_module(name='Rotate90d', module=Rotate90d)
TRANSFORMS.register_module(name='Resized', module=Resized)
TRANSFORMS.register_module(name='Affined', module=Affined)
TRANSFORMS.register_module(name='RandAffined', module=RandAffined)
TRANSFORMS.register_module(name='Rand2DElasticd', module=Rand2DElasticd)
TRANSFORMS.register_module(name='Rand3DElasticd', module=Rand3DElasticd)
TRANSFORMS.register_module(name='GridDistortiond', module=GridDistortiond)
TRANSFORMS.register_module(name='RandGridDistortiond', module=RandGridDistortiond)

# Smooth Field (Dict)
TRANSFORMS.register_module(name='RandSmoothFieldAdjustContrastd', module=RandSmoothFieldAdjustContrastd)
TRANSFORMS.register_module(name='RandSmoothFieldAdjustIntensityd', module=RandSmoothFieldAdjustIntensityd)
TRANSFORMS.register_module(name='RandSmoothDeformd', module=RandSmoothDeformd)

# Utility (Dict)
TRANSFORMS.register_module(name='Identityd', module=Identityd)
# TRANSFORMS.register_module(name='AsChannelFirstd', module=AsChannelFirstd)
TRANSFORMS.register_module(name='AsChannelLastd', module=AsChannelLastd)
# TRANSFORMS.register_module(name='AddChanneld', module=AddChanneld)
TRANSFORMS.register_module(name='EnsureChannelFirstd', module=EnsureChannelFirstd)
TRANSFORMS.register_module(name='RepeatChanneld', module=RepeatChanneld)
TRANSFORMS.register_module(name='SplitDimd', module=SplitDimd)
# TRANSFORMS.register_module(name='SplitChanneld', module=SplitChanneld)
TRANSFORMS.register_module(name='CastToTyped', module=CastToTyped)
TRANSFORMS.register_module(name='ToTensord', module=ToTensord)
TRANSFORMS.register_module(name='ToNumpyd', module=ToNumpyd)
TRANSFORMS.register_module(name='ToPIL', module=ToPIL)
TRANSFORMS.register_module(name='ToCupyd', module=ToCupyd)
TRANSFORMS.register_module(name='ToPILd', module=ToPILd)
TRANSFORMS.register_module(name='DeleteItemsd', module=DeleteItemsd)
TRANSFORMS.register_module(name='SelectItemsd', module=SelectItemsd)
TRANSFORMS.register_module(name='FlattenSubKeysd', module=FlattenSubKeysd)
TRANSFORMS.register_module(name='Transposed', module=Transposed)
TRANSFORMS.register_module(name='SqueezeDimd', module=SqueezeDimd)
TRANSFORMS.register_module(name='DataStatsd', module=DataStatsd)
TRANSFORMS.register_module(name='SimulateDelayd', module=SimulateDelayd)
TRANSFORMS.register_module(name='CopyItemsd', module=CopyItemsd)
TRANSFORMS.register_module(name='ConcatItemsd', module=ConcatItemsd)
TRANSFORMS.register_module(name='Lambdad', module=Lambdad)
TRANSFORMS.register_module(name='RandLambdad', module=RandLambdad)
TRANSFORMS.register_module(name='RemoveRepeatedChanneld', module=RemoveRepeatedChanneld)
TRANSFORMS.register_module(name='LabelToMaskd', module=LabelToMaskd)
TRANSFORMS.register_module(name='FgBgToIndicesd', module=FgBgToIndicesd)
TRANSFORMS.register_module(name='ClassesToIndicesd', module=ClassesToIndicesd)
TRANSFORMS.register_module(name='ConvertToMultiChannelBasedOnBratsClassesd',
                           module=ConvertToMultiChannelBasedOnBratsClassesd)
TRANSFORMS.register_module(name='AddExtremePointsChanneld', module=AddExtremePointsChanneld)
TRANSFORMS.register_module(name='TorchVisiond', module=TorchVisiond)
TRANSFORMS.register_module(name='RandTorchVisiond', module=RandTorchVisiond)
TRANSFORMS.register_module(name='MapLabelValued', module=MapLabelValued)
TRANSFORMS.register_module(name='EnsureTyped', module=EnsureTyped)
TRANSFORMS.register_module(name='IntensityStatsd', module=IntensityStatsd)
TRANSFORMS.register_module(name='ToDeviced', module=ToDeviced)
TRANSFORMS.register_module(name='CuCIMd', module=CuCIMd)
TRANSFORMS.register_module(name='RandCuCIMd', module=RandCuCIMd)
TRANSFORMS.register_module(name='AddCoordinateChannelsd', module=AddCoordinateChannelsd)

# MetaTensor
TRANSFORMS.register_module(name='ToMetaTensord', module=ToMetaTensord)
TRANSFORMS.register_module(name='FromMetaTensord', module=FromMetaTensord)



import os

from monai.data import Dataset
from sklearn.model_selection import train_test_split

from medlab.registry import DATASETS, TRANSFORMS


@DATASETS.register_module()
class BaseClsDataset(Dataset):
    META_INFO = {'classes': list()}

    def __init__(
            self,
            data_root: str,
            data_suffix: str,
            subset: str = None,
            test_size: float = 0.2,
            seed: int = 1234,
            transforms: dict = None,
            annotation_file: str = None,
    ):
        """
        base classification dataset
        if annotation_file is not None, use annotation file to load data
        else, use subset and test_size to split data

        if subset is None, use all data
        if subset is 'train', use train data
        if subset is 'val', use val data

        :param data_root: data root path
        :param data_suffix: image suffix
        :param subset: train or val or None
        :param test_size: test size
        :param seed: random seed
        :param transforms: MONAI data transforms cfg
        :param annotation_file: annotation file path
        """
        assert subset in ['train', 'val', None]

        classes = sorted([i for i in os.listdir(data_root) if os.path.isdir(os.path.join(data_root, i))])
        self.META_INFO['classes'] = classes
        images = []
        labels = []
        for root, dirs, files in os.walk(data_root):
            for file in files:
                if file.endswith(data_suffix):
                    images.append(os.path.join(root, file))
                    labels.append(classes.index(os.path.basename(root)))

        assert len(images) > 0, 'No images found in {}'.format(data_root)
        assert len(images) == len(labels), 'The number of images:{} and labels:{} should be equal'.format(len(images),
                                                                                                          len(labels))

        if subset is not None:
            train_x, val_x, train_y, val_y = train_test_split(images, labels, test_size=test_size, random_state=seed,
                                                              stratify=labels)
            if subset == 'train':
                images = train_x
                labels = train_y
            else:
                images = val_x
                labels = val_y
        data = [{'images': i, 'labels': m} for i, m in zip(images, labels)]
        transforms = TRANSFORMS.build(transforms)
        super().__init__(data=data, transform=transforms)

    @property
    def classes(self):
        return self.META_INFO['classes']



from torchvision.models._api import list_models, get_model_builder

from medlab.registry import MODELS

# for model_name in list_models():
#     MODELS.register_module(name=model_name + '_torchvision', module=get_model_builder(model_name))



import copy
from typing import Union, Dict, Any

import lightning.pytorch as pl
from mmengine.config import Config, ConfigDict
from monai.data import DataLoader
from torch import nn

from medlab.registry import DATASETS, MODELS, OPTIMIZERS, METRICS, LOSSES, LR_SCHEDULERS


class BaseTask(pl.LightningModule):
    def __init__(
            self,
            model: Union[nn.Module, Dict],
            train_dataloader: Union[DataLoader, Dict] = None,
            val_dataloader: Union[DataLoader, Dict] = None,
            test_dataloader: Union[DataLoader, Dict] = None,
            predict_dataloader: Union[DataLoader, Dict] = None,
            loss_func: Union[nn.Module, Dict] = None,
            optims: Any = None,
            metrics: Any = None,
            train_cfg: Dict = None,
            val_cfg: Dict = None,
            test_cfg: Dict = None,
            **kwargs
    ):
        """
        BaseTask
        :param model: nn.Module or model cfg
        :param train_dataloader: train dataloader or dataloader cfg
        :param val_dataloader: val dataloader or dataloader cfg
        :param test_dataloader: test dataloader or dataloader cfg
        :param loss_func: loss function or loss function cfg
        :param optims: PyTorch-lightning configure_optimizers
        :param metrics: metrics cfg
        :param train_cfg: train cfg
        :param val_cfg: val cfg
        :param test_cfg: test cfg
        :param kwargs: reserved parameters
        """
        super().__init__()

        self._model = self.build_model(model)
        self._train_dataloader = self.build_dataloader(train_dataloader)
        self._val_dataloader = self.build_dataloader(val_dataloader)
        self._test_dataloader = self.build_dataloader(test_dataloader)
        self._predict_dataloader = self.build_dataloader(predict_dataloader)
        self.loss_func = self.build_loss_func(loss_func)
        self.optims = optims
        self.metrics_key, self.metrics = self.build_metric(metrics)
        self.train_cfg = train_cfg if train_cfg is not None else dict()
        self.val_cfg = val_cfg if val_cfg is not None else dict()
        self.test_cfg = test_cfg if test_cfg is not None else dict()

    def train_dataloader(self):
        """
        :return: train dataloader (PyTorch-lightning needs)
        """
        return self._train_dataloader

    def val_dataloader(self):
        """
        :return: val dataloader (PyTorch-lightning needs)
        """
        return self._val_dataloader

    def test_dataloader(self):
        """
        :return: test dataloader (PyTorch-lightning needs)
        """
        return self._test_dataloader
    
    def predict_dataloader(self):
        """
        :return: predict dataloader (PyTorch-lightning needs)
        """
        return self._predict_dataloader

    @staticmethod
    def build_dataloader(dataloader: Union[DataLoader, Dict]) -> DataLoader:
        """
        build dataloader from cfg
        :param dataloader: dataloader or dataloader cfg
        :return: MONAI dataloader
        """
        if isinstance(dataloader, (DataLoader, type(None))):
            return dataloader

        dataloader_cfg = copy.deepcopy(dataloader)
        dataset_cfg = dataloader_cfg.pop('dataset')
        if isinstance(dataset_cfg, Dict):
            dataset = DATASETS.build(dataset_cfg)
        else:
            dataset = dataset_cfg

        data_loader = DataLoader(
            dataset=dataset,
            **dataloader_cfg
        )

        return data_loader

    @staticmethod
    def build_model(model: Union[nn.Module, Dict]) -> nn.Module:
        """
        build model from cfg
        :param model: nn.Module or model cfg
        :return: nn.Module
        """
        if isinstance(model, nn.Module):
            return model
        return MODELS.build(model)

    @staticmethod
    def build_loss_func(loss_func: Union[nn.Module, Dict]) -> nn.Module:
        """
        build loss function from cfg
        :param loss_func: nn.Module or loss cfg
        :return: nn.Module loss
        """
        if isinstance(loss_func, nn.Module):
            return loss_func

        loss_cfg = copy.deepcopy(loss_func)
        if loss_cfg.get('type', None) == 'DeepSupervisionLoss':
            loss = loss_cfg.get('loss')
            if isinstance(loss, (dict, ConfigDict, Config)):
                loss = LOSSES.build(loss)
                loss_cfg.update(dict(loss=loss))
        return LOSSES.build(loss_func)

    @staticmethod
    def build_optims(optims: Any, params: Any) -> Any:
        """
        build optimizer and lr schedules from cfg
        :param optims: optimizer and lr scheduler cfg
        :param params: model parameters to be optimized
        :return: optimizer and lr scheduler dict (PyTorch-lightning needs)
        """
        optims_cfg = copy.deepcopy(optims)
        optimizer = optims_cfg.get('optimizer', None)
        lr_scheduler = optims_cfg.get('lr_scheduler', None)
        assert optimizer is not None, 'optimizer must be provided'
        optimizer.update(dict(params=params))
        optimizer = OPTIMIZERS.build(optimizer)

        if lr_scheduler is not None:
            scheduler = lr_scheduler.pop('scheduler', None)
            scheduler.update(dict(optimizer=optimizer))
            scheduler = LR_SCHEDULERS.build(scheduler)
            lr_scheduler.update(dict(scheduler=scheduler))
            optims_cfg.update(dict(optimizer=optimizer, lr_scheduler=lr_scheduler))
            return optims_cfg
        return optimizer

    @staticmethod
    def build_metric(metrics: Any) -> Any:
        """
        build metrics from cfg
        :param metrics: metrics cfg
        :return: MONAI metrics
        """
        return METRICS.build(metrics)

    def configure_optimizers(self):
        """
        :return: optimizer and lr scheduler dict (PyTorch-lightning needs)
        """
        return self.build_optims(self.optims, self.parameters())



import copy
import os.path as osp
import time
from typing import Dict, Union

from lightning.pytorch.trainer import Trainer as LightningTrainer
from mmengine.config import Config, ConfigDict

from medlab.registry import TRAINERS, CALLBACKS, LOGGERS

ConfigType = Union[Dict, Config, ConfigDict]


@TRAINERS.register_module()
class Trainer(LightningTrainer):
    def __init__(self, prefix=None, *args, **kwargs):
        """
        build loggers and callbacks, get time prefix, init lightning trainer
        :param args: same as lightning trainer
        :param kwargs: same as lightning trainer
        """
        time_prefix = time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime()) if prefix is None else prefix
        logger = kwargs.get('logger', None)
        callbacks = kwargs.get('callbacks', None)
        if logger is not None:
            logger = self.build_loggers(logger, kwargs.get('default_root_dir', None), time_prefix)
        if callbacks is not None:
            callbacks = self.build_callbacks(kwargs.get('callbacks', None), kwargs.get('default_root_dir', None),
                                             time_prefix)
        kwargs.update(dict(logger=logger, callbacks=callbacks))
        super().__init__(*args, **kwargs)
        self._time_prefix = time_prefix

    @staticmethod
    def build_loggers(logger, work_dir, name):
        """
        build logger(s) from cfg
        :param logger: logger cfg
        :param work_dir: logger save dir
        :param name: logger name
        :return: lightning logger(s)
        """
        loggers_cfg = copy.deepcopy(logger)

        if isinstance(loggers_cfg, list):
            for i in loggers_cfg:
                i.update(dict(save_dir=work_dir, name=name))
        else:
            loggers_cfg.update(dict(save_dir=work_dir, name=name))

        return LOGGERS.build(loggers_cfg)

    @staticmethod
    def build_callbacks(callbacks, work_dir, name):
        """
        build callback(s) from cfg
        :param callbacks: callback cfg
        :param work_dir: if ModelCheckpoint in callbacks , save dir is work_dir/name/checkpoints
        :param name: checkpoint dir name
        :return:
        """
        callbacks_cfg = copy.deepcopy(callbacks)
        if isinstance(callbacks_cfg, list):
            for i in callbacks_cfg:
                if i.get('type', None) == 'ModelCheckpoint':
                    i.update(dict(dirpath=osp.join(work_dir, name, 'checkpoints')))
        else:
            if callbacks_cfg.get('type', None) == 'ModelCheckpoint':
                callbacks_cfg.update(dict(dirpath=osp.join(work_dir, name, 'checkpoints')))
        return CALLBACKS.build(callbacks_cfg)

    @classmethod
    def from_cfg(cls, cfg: ConfigType):
        """
        build trainer from cfg
        :param cfg:
        :return: trainer
        """
        cfg = copy.deepcopy(cfg)
        trainer = cls(
            accelerator=cfg.get('accelerator', 'auto'),
            strategy=cfg.get('strategy', 'auto'),
            devices=cfg.get('devices', 'auto'),
            num_nodes=cfg.get('num_nodes', 1),
            precision=cfg.get('precision', "32-true"),
            logger=cfg.get('logger', None),
            callbacks=cfg.get('callbacks', None),
            fast_dev_run=cfg.get('fast_dev_run', False),
            max_epochs=cfg.get('max_epochs', None),
            min_epochs=cfg.get('min_epochs', None),
            max_steps=cfg.get('max_steps', -1),
            min_steps=cfg.get('min_steps', None),
            max_time=cfg.get('max_time', None),
            limit_train_batches=cfg.get('limit_train_batches', None),
            limit_val_batches=cfg.get('limit_val_batches', None),
            limit_test_batches=cfg.get('limit_test_batches', None),
            limit_predict_batches=cfg.get('limit_predict_batches', None),
            overfit_batches=cfg.get('overfit_batches', 0.0),
            val_check_interval=cfg.get('val_check_interval', None),
            check_val_every_n_epoch=cfg.get('check_val_every_n_epoch', 1),
            num_sanity_val_steps=cfg.get('num_sanity_val_steps', None),
            log_every_n_steps=cfg.get('log_every_n_steps', None),
            enable_checkpointing=cfg.get('enable_checkpointing', None),
            enable_progress_bar=cfg.get('enable_progress_bar', None),
            enable_model_summary=cfg.get('enable_model_summary', None),
            accumulate_grad_batches=cfg.get('accumulate_grad_batches', 1),
            gradient_clip_val=cfg.get('gradient_clip_val', None),
            gradient_clip_algorithm=cfg.get('gradient_clip_algorithm', None),
            deterministic=cfg.get('deterministic', None),
            benchmark=cfg.get('benchmark', None),
            inference_mode=cfg.get('inference_mode', True),
            use_distributed_sampler=cfg.get('use_distributed_sampler', True),
            profiler=cfg.get('profiler', None),
            detect_anomaly=cfg.get('detect_anomaly', False),
            barebones=cfg.get('barebones', False),
            plugins=cfg.get('plugins', None),
            sync_batchnorm=cfg.get('sync_batchnorm', False),
            reload_dataloaders_every_n_epochs=cfg.get('reload_dataloaders_every_n_epochs', 0),
            default_root_dir=cfg.get('default_root_dir', None)
        )
        return trainer

    @property
    def time_prefix(self):
        """
        get time prefix
        :return: string time prefix
        """
        return self._time_prefix



from lightning.pytorch.callbacks import ModelCheckpoint, LearningRateMonitor, EarlyStopping

from medlab.registry import CALLBACKS

CALLBACKS.register_module(name='ModelCheckpoint', module=ModelCheckpoint)
CALLBACKS.register_module(name='LearningRateMonitor', module=LearningRateMonitor)
CALLBACKS.register_module(name='EarlyStopping', module=EarlyStopping)



import os.path as osp

from monai.data import decollate_batch

from medlab.registry import TASKS, INFERERS, TRANSFORMS
from .base_task import BaseTask
from typing import Any, OrderedDict, Dict
import copy
import torch
import pandas as pd
import os
import json
import pathlib
from copy import deepcopy
@TASKS.register_module()
class BaseSegTask(BaseTask):
    def __init__(self, *args, **kwargs):
        """
        base segmentation task, one input, one output
        """
        super().__init__(*args, **kwargs)
        self.val_inferer = INFERERS.build(self.val_cfg.get('inferer', dict(type='SimpleInferer')))
        self.test_inferer = INFERERS.build(self.test_cfg.get('inferer', dict(type='SimpleInferer')))
        num_classes = kwargs.get('num_classes', None)
        assert num_classes is not None, "num_classes must be specified in model"
        if num_classes == 1:
            self.post_pred = TRANSFORMS.build([
                dict(type='Activations', sigmoid=True),
                dict(type='AsDiscrete', threshold=0.5)
            ])
            self.post_label = TRANSFORMS.build([])
            post_save = [
                dict(type='ToDevice', device='cpu'),
            ]
        else:
            self.post_pred = TRANSFORMS.build([dict(type='AsDiscrete', argmax=True, to_onehot=num_classes)])
            self.post_label = TRANSFORMS.build([dict(type='AsDiscrete', to_onehot=num_classes)])
            post_save = [
                dict(type='AsDiscrete', argmax=True),
                dict(type='ToDevice', device='cpu')
            ]
        save_dir = self.test_cfg.get('save_dir', None)
        pred_save = None
        if save_dir is not None:
            post_save.append(
                dict(
                    type='SaveImage',
                    output_dir=osp.join(self.test_cfg.get('save_dir', None), 'test_result'),
                    resample=True,
                    output_postfix='',
                    output_ext=self.test_cfg.get('output_ext', '.png'),
                    scale=self.test_cfg.get('scale', 255.0),
                    squeeze_end_dims=True,
                    separate_folder=False
                )
            )
            pred_save = deepcopy(post_save)
            pred_save[-1]['output_dir'] = osp.join(self.test_cfg.get('save_dir', None), 'predict_result')

        self.post_save = TRANSFORMS.build(post_save) if post_save is not None else None
        self.pred_save = TRANSFORMS.build(pred_save) if pred_save is not None else None
        self.train_metrics_key = copy.deepcopy(self.metrics_key)
        self.train_metrics = copy.deepcopy(self.metrics)

        self.val_metrics_key = copy.deepcopy(self.metrics_key)
        self.val_metrics = copy.deepcopy(self.metrics)

        self.training_step_loss = []
        self.validation_step_loss = []

        self.test_names = []
        self.test_dice = []
        self.test_acc = []

    def forward(self, x):
        return self.parse_outputs(self._model(x))

    def training_step(self, batch, batch_idx):
        """
        training step for classification task
        :param batch: batch data
        :param batch_idx: batch index
        :return: loss
        """
        images = batch['images']
        labels = batch['labels']

        batch_size = images.shape[0]
        outputs = self.forward(images)
        
        if self.loss_func.__class__.__name__ == 'CrossEntropyLoss':
            labels = labels.squeeze(1).long()
        loss = self.loss_func(outputs, labels)

        if self.loss_func.__class__.__name__ == 'CrossEntropyLoss':
            labels = labels.unsqueeze(1).long()

        if isinstance(outputs, (tuple, list)):
            outputs = outputs[0]

        outputs = [self.post_pred(i) for i in decollate_batch(outputs)]
        labels = [self.post_label(i) for i in decollate_batch(labels)]


        if len(self.train_metrics) > 0:
            for metric in self.train_metrics:
                metric(outputs, labels)
        
        self.training_step_loss.append(loss)
        # self.log("train_step_loss", loss, sync_dist=True, batch_size=batch_size)
        
        
        return loss

    def on_train_epoch_end(self):
        self.log_dict(self.parse_train_metrics(), sync_dist=True)
        epoch_mean = torch.stack(self.training_step_loss).mean()
        self.log("train_loss", epoch_mean.item(), sync_dist=True)
        self.training_step_loss.clear()

    def validation_step(self, batch, batch_idx):
        """
        validation step for classification task
        :param batch: batch data
        :param batch_idx: batch index
        :return: None
        """
        images = batch['images']
        labels = batch['labels']

        batch_size = images.shape[0]
        # outputs = self.forward(images)
        outputs = self.val_inferer(inputs=images, network=self.forward)

        if self.loss_func.__class__.__name__ == 'CrossEntropyLoss':
            labels = labels.squeeze(1).long()
        loss = self.loss_func(outputs, labels)
        if self.loss_func.__class__.__name__ == 'CrossEntropyLoss':
            labels = labels.unsqueeze(1).long()

        if isinstance(outputs, (tuple, list)):
            outputs = outputs[0]

        outputs = [self.post_pred(i) for i in decollate_batch(outputs)]
        labels = [self.post_label(i) for i in decollate_batch(labels)]

        if len(self.val_metrics) > 0:
            for metric in self.val_metrics:
                metric(outputs, labels)
        self.validation_step_loss.append(loss)

        # self.log('val_step_loss', loss.item(), sync_dist=True, batch_size=batch_size)

    def on_validation_epoch_end(self):
        """
        validation epoch end hook, parse and log metrics
        """
        self.log_dict(self.parse_val_metrics(), sync_dist=True)
        epoch_mean = torch.stack(self.validation_step_loss).mean()
        self.log("val_loss", epoch_mean.item(), sync_dist=True)
        self.validation_step_loss.clear()
        # self.log_dict(self.parse_metrics(), sync_dist=True)

    def test_step(self, batch, batch_idx):
        """
        test step for classification task, save predictions to disk
        :param batch: batch data
        :param batch_idx: batch index
        :return:
        """
        images = batch["images"]
        labels = batch["labels"]

        outputs = self.test_inferer(inputs=images, network=self.forward)

        if isinstance(outputs, (tuple, list)):
            outputs = outputs[0]

        outputs = [self.post_pred(i) for i in decollate_batch(outputs)]
        labels = [self.post_label(i) for i in decollate_batch(labels)]

        for metric in self.metrics:
            metric(outputs, labels)

        if len(self.val_metrics) > 0:
            for metric in self.val_metrics:
                metric(outputs, labels)

        # outputs[0].meta['filename_or_obj'] = batch['images_meta_dict']['filename_or_obj'][0]

        # print(self.post_save(outputs[0]))
        for i in outputs:
            self.post_save(i)
            self.test_names.append(i.meta['filename_or_obj'])


    def on_test_epoch_end(self):
        """
        test epoch end hook, parse metrics
        """
        for i in self.val_metrics:
            if i.__class__.__name__ == "DiceMetric":
                # 展开成一维
                i.reduction = "none"
                self.test_dice = i.aggregate().cpu().numpy().ravel().tolist()
                
            if i.__class__.__name__ == "ConfusionMatrixMetric":
                i.reduction = "none"
                self.test_acc = i.aggregate()[0].cpu().numpy().ravel().tolist()
        
        df = pd.DataFrame({'name': self.test_names, 'dice': self.test_dice, 'accuracy': self.test_acc})
        file_path = pathlib.Path(df['name'][0])
        base_dir = file_path.parent.parent
        df['label'] = [os.path.join(base_dir, 'Segmentation', pathlib.Path(i).name) for i in df['name']]
        save_dir = self.test_cfg.get('save_dir', None)
        df['pred'] = [os.path.join(save_dir, 'test_result', pathlib.Path(i).name) for i in df['name']]
        
        df.to_csv(os.path.join(save_dir, "test.csv"), index=False, encoding='utf-8')
        dice = df['dice'].mean()
        acc = df['accuracy'].mean()
        metric = {'dice': dice, 'accuracy': acc}
        json_path = os.path.join(save_dir, "metric.json")
        with open(json_path, 'w') as f:
            json.dump(metric, f)

    def parse_train_metrics(self):
        """
        parse metrics to dict
        :return: metrics dict
        """
        values = []
        for metric in self.train_metrics:
            value = metric.aggregate()

            if isinstance(value, list):
                values.extend([v.item() for v in value])
            else:
                values.append(value.item())

        value_dict = dict(zip(self.train_metrics_key, values))
        for metric in self.train_metrics:
            metric.reset()
        value_dict = {f'train_{k}': v for k, v in value_dict.items()}
        return value_dict

    def predict_step(self, batch, batch_idx, dataloader_idx=0) -> Any:
        images = batch["images"]
        outputs = self.test_inferer(inputs=images, network=self.forward)
        if isinstance(outputs, (tuple, list)):
            outputs = outputs[0]
        outputs = [self.post_pred(i) for i in decollate_batch(outputs)]
        batch["preds"] = outputs
        
        for i in outputs:
            self.pred_save(i)

    def parse_val_metrics(self):
        """
        parse metrics to dict
        :return: metrics dict
        """
        values = []
        for metric in self.val_metrics:
            value = metric.aggregate()

            if isinstance(value, list):
                values.extend([v.item() for v in value])
            else:
                values.append(value.item())

        value_dict = dict(zip(self.val_metrics_key, values))
        for metric in self.val_metrics:
            metric.reset()
        value_dict = {f'val_{k}': v for k, v in value_dict.items()}
        return value_dict

    @staticmethod
    def parse_outputs(outputs):
        """
        parse outputs to tensor or list
        :return: outputs dict
        """
        if isinstance(outputs, (OrderedDict, Dict)):
            outputs = list(outputs.values())
            if len(outputs) == 1:
                outputs = outputs[0]
        return outputs



import argparse
import logging
import os.path as osp
import warnings
import os
from monai.data import DataLoader, Dataset

from lightning.pytorch import seed_everything
from mmengine.config import Config

from medlab.registry import TRAINERS, TASKS, TRANSFORMS


def parse_args():
    parser = argparse.ArgumentParser(description='Train a segmentor')
    parser.add_argument('checkpoint', help='model checkpoint file path')
    parser.add_argument('path', help='predict file or dir path of filelist')
    parser.add_argument('--save-dir', help='the dir to save logs and models')
    args = parser.parse_args()
    return args

def get_predict_files(path):
    try:
        path = eval(path)
    except:
        path = str(path)
        

    files = []
    if isinstance(path, list):
        files = path
    else:
        if os.path.isdir(path):
            for file in os.listdir(path):
                filename = os.path.join(path, file)
                if os.path.isfile(filename):
                    files.append(filename)
        if os.path.isfile(path):
            files.append(path)
    files = sorted(files)
    files = [{'images': i} for i in files]
    return files


def predict():
    warnings.filterwarnings("ignore")
    logging.disable(logging.WARNING)
    args = parse_args()
    assert args.path
    cfg_path = osp.join(osp.dirname(osp.dirname(args.checkpoint)), 'config.py')

    assert osp.exists(cfg_path)

    cfg = Config.fromfile(cfg_path)

    seed = cfg.get('seed', 1)
    ckpt_path = cfg.get('ckpt_path', None)
    trainer_cfg = cfg.get('trainer', dict())
    seed_everything(seed)
    
    if args.checkpoint is not None:
        ckpt_path = args.checkpoint

    if args.save_dir is not None:
        # update configs according to CLI args if args.work_dir is not None
        save_dir = args.save_dir
    else:
        save_dir = cfg.get('save_dir', osp.join(osp.dirname(ckpt_path), osp.splitext(osp.basename(ckpt_path))[0]))

    trainer_cfg['default_root_dir'] = save_dir

    trainer = TRAINERS.build(trainer_cfg)

    test_cfg = cfg.get('test_cfg', dict())
    test_cfg.update(dict(save_dir=save_dir))

    
    files = get_predict_files(args.path)
    predict_transforms = cfg.val_dataloader.dataset.transforms
    new_transforms = []
    for d in predict_transforms:
        for key in ['keys', 'mode']:
            if key in d and isinstance(d[key], list):
                d[key] = d[key][0]
        if d['keys'] == 'labels':
            continue
        new_transforms.append(d)
    print(new_transforms)
    new_transforms = TRANSFORMS.build(new_transforms)
    predict_dataset = Dataset(data=files, transform=new_transforms)
    predict_dataloader = DataLoader(dataset=predict_dataset, batch_size=1, num_workers=0)


    task = TASKS.build(
        dict(
            type=cfg.get('task_type', None),
            model=cfg.get('model', None),
            train_dataloader=None,
            val_dataloader=None,
            test_dataloader=None,
            loss_func=cfg.get('loss_func', None),
            optims=None,
            metrics=cfg.get('metrics', None),
            train_cfg=dict(),
            val_cfg=dict(),
            test_cfg=test_cfg,
            num_classes=cfg.get('num_classes', None)
        )
    )
    trainer.predict(task, dataloaders=predict_dataloader, ckpt_path=ckpt_path)


if __name__ == '__main__':
    predict()



from mmengine.registry import Registry

from .build_functions import build_model_from_cfg, build_logger_from_cfg, build_callback_from_cfg, \
    build_transform_from_cfg, build_metric_from_cfg

# manage PyTorch-lightning Trainer
TRAINERS = Registry('trainer', locations=['medlab.core'], scope='medlab')
# manage PyTorch-lightning Callbacks
CALLBACKS = Registry('callback', build_func=build_callback_from_cfg, locations=['medlab.core.callbacks'], scope='medlab')
# manage PyTorch-lightning loggers
LOGGERS = Registry('logger', build_func=build_logger_from_cfg, locations=['medlab.core.loggers'], scope='medlab')
# manage different vision tasks based on LightningModule
TASKS = Registry('task', locations=['medlab.tasks'], scope='medlab')

# manage datasets
DATASETS = Registry('dataset', locations=['medlab.datasets'], scope='medlab')
# manage MONAI transforms
TRANSFORMS = Registry('transform', build_func=build_transform_from_cfg, locations=['medlab.core.transforms'], scope='medlab')
# manage MONAI inferers
INFERERS = Registry('inferer', locations=['medlab.core.inferers'], scope='medlab')

# manage PyTorch models
MODELS = Registry('model', build_func=build_model_from_cfg, locations=['medlab.models'], scope='medlab')
LOSSES = Registry('loss', locations=['medlab.core.losses'], scope='medlab')
OPTIMIZERS = Registry('optimizer', locations=['medlab.core.optimizers'], scope='medlab')
LR_SCHEDULERS = Registry('learning rate scheduler', locations=['medlab.core.lr_schedulers'], scope='medlab')
METRICS = Registry('metric', build_func=build_metric_from_cfg, locations=['medlab.core.metrics'], scope='medlab')



from monai.losses import DiceLoss, DiceCELoss, DeepSupervisionLoss
from torch.nn import CrossEntropyLoss, BCEWithLogitsLoss, MSELoss, L1Loss, BCELoss

from medlab.registry import LOSSES

LOSSES.register_module(name='BCELoss', module=BCELoss)
LOSSES.register_module(name='CrossEntropyLoss', module=CrossEntropyLoss)
LOSSES.register_module(name='BCEWithLogitsLoss', module=BCEWithLogitsLoss)
LOSSES.register_module(name='MSELoss', module=MSELoss)
LOSSES.register_module(name='L1Loss', module=L1Loss)
LOSSES.register_module(name='DiceLoss', module=DiceLoss)
LOSSES.register_module(name='DiceCELoss', module=DiceCELoss)
LOSSES.register_module(name='DeepSupervisionLoss', module=DeepSupervisionLoss)



from .callbacks import *
from .inferers import *
from .loggers import *
from .losses import *
from .lr_schedulers import *
from .metrics import *
from .optimizers import *
from .transforms import *
from .trainer import *


import os
import argparse
import shutil
from medlab.script import train, test, predict

def parse_args():
    # parse arguments
    parser = argparse.ArgumentParser(description='medlab')
    parser.add_argument('--init-path', type=str, default='./', help='config file path')
    return parser.parse_args()

def init_config():
    args  = parse_args()
    current_dir = os.path.dirname(os.path.abspath(__file__))
    source_cfg_dir = os.path.join(current_dir, 'configs')
    target_cfg_dir = os.path.join(args.init_path, 'configs')
    shutil.copytree(source_cfg_dir, target_cfg_dir)

def init_enviroment():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    requirements = os.path.join(os.path.dirname(current_dir), 'requirements.txt')
    os.system('pip install -r {}'.format(requirements))


from .build_functions import build_model_from_cfg, build_logger_from_cfg, build_callback_from_cfg, \
    build_transform_from_cfg, build_metric_from_cfg
from .registry import TRAINERS, DATASETS, TRANSFORMS, MODELS, LOSSES, OPTIMIZERS, LR_SCHEDULERS, METRICS, INFERERS, \
    CALLBACKS, TASKS, LOGGERS



from setuptools import setup, find_packages

# with open('readme.md', 'r', encoding='utf-8') as fh:
#     long_description = fh.read()
long_description = ""

setup(
    name='medlab',
    version='2024.01.31',
    author='yjiang',
    author_email='1900812907@qq.com',
    description='medical deep learning toolkit',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://gitee.com/Eason596/py-package-release-test',
    packages=find_packages(),
    include_package_data=True,
    package_data={"medlab": ["testing/*", "configs/*","requirements.txt"]},
    install_requires=[],
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'med-init-env = medlab:init_enviroment',
            'med-init-config = medlab:init_config',
            'med-train = medlab:train',
            'med-test = medlab:test',
            'med-predict = medlab:predict'
        ]
    }
)


from typing import Optional, Union

import torch.nn as nn
from mmengine.config import Config, ConfigDict
from mmengine.registry import Registry, build_from_cfg
from monai.transforms import Compose
from monai.data import PILReader

def build_model_from_cfg(
        cfg: Union[dict, ConfigDict, Config],
        registry: Registry,
        default_args: Optional[Union[dict, 'ConfigDict', 'Config']] = None
) -> 'nn.Module':
    """
    Build a PyTorch model from config dict(s). Different from
    ``build_from_cfg``, if cfg is a list, a ``nn.Sequential`` will be built.
    :param cfg: model config
    :param registry: model registry
    :param default_args: default arguments for the build function
    :return: nn.Module
    """
    if isinstance(cfg, list):
        modules = [build_from_cfg(_cfg, registry, default_args) for _cfg in cfg]
        return nn.Sequential(*modules)
    else:
        return build_from_cfg(cfg, registry, default_args)


def build_logger_from_cfg(
        cfg: Union[dict, ConfigDict, Config],
        registry: Registry,
        default_args: Optional[Union[dict, 'ConfigDict', 'Config']] = None
):
    """
    Build loggers from config dict(s).
    :param cfg: logger config
    :param registry: logger registry
    :param default_args: default arguments for the build function
    :return: PyTorch-lightning logger(s)
    """
    if isinstance(cfg, list):
        loggers = []
        for _cfg in cfg:
            loggers.append(build_from_cfg(_cfg, registry, default_args))
        return loggers
    else:
        return build_from_cfg(cfg, registry, default_args)


def build_callback_from_cfg(
        cfg: Union[dict, ConfigDict, Config],
        registry: Registry,
        default_args: Optional[Union[dict, 'ConfigDict', 'Config']] = None
):
    """
    Build callbacks from config dict(s).
    :param cfg: callback config
    :param registry: callback registry
    :param default_args: default arguments for the build function
    :return: pytorch-lightning callback(s)
    """
    if isinstance(cfg, list):
        callbacks = [build_from_cfg(_cfg, registry, default_args) for _cfg in cfg]
        return callbacks
    else:
        return build_from_cfg(cfg, registry, default_args)


def build_transform_from_cfg(
        cfg: Union[dict, ConfigDict, Config],
        registry: Registry,
        default_args: Optional[Union[dict, 'ConfigDict', 'Config']] = None
):
    """
    Build transforms from config dict(s).
    :param cfg: transform config
    :param registry: transform registry
    :param default_args: default arguments for the build function
    :return: Compose transform(s) from MONAI
    """
    if isinstance(cfg, list):
        transforms = [
            build_transform_from_cfg(_cfg, registry, default_args) for _cfg in cfg
        ]
        return Compose(transforms)
    else:
        if cfg is not None and (cfg.get('type', None) == 'LoadImaged' or cfg.get('type', None) == 'LoadImage'):
            if cfg.get('reader', None) == 'PILReader':
                color = cfg.pop('color', None)
                assert color in color, "color参数错误"
                cfg['converter'] = lambda x: x.convert(color)
        return build_from_cfg(cfg, registry, default_args)


def build_metric_from_cfg(
        cfg: Union[dict, ConfigDict, Config],
        registry: Registry,
        default_args: Optional[Union[dict, 'ConfigDict', 'Config']] = None
):
    """
    Build metrics from config dict(s).
    :param cfg: metric config
    :param registry: metric registry
    :param default_args: default arguments for the build function
    :return: Metric name(s) and metric(s)
    """
    keys = []
    metrics = []

    def get_metric_key(args):
        if args.get('type') == 'ConfusionMatrixMetric':
            key = args.get('metric_name')
            if isinstance(key, list):
                return key
            else:
                return [key]
        else:
            return [args.get('type')]

    if isinstance(cfg, list):
        for _cfg in cfg:
            metrics.append(build_from_cfg(_cfg, registry, default_args))
            keys.extend(get_metric_key(_cfg))
        return keys, metrics
    else:
        return get_metric_key(cfg), [build_from_cfg(cfg, registry, default_args)]



from timm import create_model, list_models

from medlab.registry import MODELS

# for i in [i for i in list_models() if "tf_" not in i]:
#     MODELS.register_module(name='cls_'+ i, module=create_model)



from monai.networks.nets import UNet, VNet, SwinUNETR, UNETR, AHNet, BasicUNetPlusPlus, AttentionUnet, RegUNet, \
    LocalNet, SegResNet
from monai.networks.nets import ResNet, resnet10, resnet18, resnet34, resnet50, resnet101, resnet152, DenseNet, densenet121, densenet169, \
    densenet201, densenet264, EfficientNetBN, ViT, SENet, senet154, seresnet50, seresnet101, seresnet152, seresnext50, seresnext101

from medlab.registry import MODELS

MODELS.register_module(name='resnet_monai', module=ResNet)
MODELS.register_module(name='resnet10_monai', module=resnet10)
MODELS.register_module(name='resnet18_monai', module=resnet18)
MODELS.register_module(name='resnet34_monai', module=resnet34)
MODELS.register_module(name='resnet50_monai', module=resnet50)
MODELS.register_module(name='resnet101_monai', module=resnet101)
MODELS.register_module(name='resnet152_monai', module=resnet152)
MODELS.register_module(name='densenet_monai', module=DenseNet)
MODELS.register_module(name='densenet121_monai', module=densenet121)
MODELS.register_module(name='densenet169_monai', module=densenet169)
MODELS.register_module(name='densenet201_monai', module=densenet201)
MODELS.register_module(name='densenet264_monai', module=densenet264)
MODELS.register_module(name='senet_monai', module=SENet)
MODELS.register_module(name='senet154_monai', module=senet154)
MODELS.register_module(name='seresnet50_monai', module=seresnet50)
MODELS.register_module(name='seresnet101_monai', module=seresnet101)
MODELS.register_module(name='seresnet152_monai', module=seresnet152)
MODELS.register_module(name='seresnext50_monai', module=seresnext50)
MODELS.register_module(name='seresnext101_monai', module=seresnext101)
MODELS.register_module(name='efficientnet_monai', module=EfficientNetBN)
MODELS.register_module(name='vit_monai', module=ViT)

MODELS.register_module(name='unet_monai', module=UNet)
MODELS.register_module(name='vnet_monai', module=VNet)
MODELS.register_module(name='swinunetr_monai', module=SwinUNETR)
MODELS.register_module(name='unetr_monai', module=UNETR)
MODELS.register_module(name='ahnet_monai', module=AHNet)
MODELS.register_module(name='unetplusplus_monai', module=BasicUNetPlusPlus)
MODELS.register_module(name='attentionunet_monai', module=AttentionUnet)
MODELS.register_module(name='regunet_monai', module=RegUNet)
MODELS.register_module(name='localnet_monai', module=LocalNet)
MODELS.register_module(name='segresnet_monai', module=SegResNet)



from monai.data import Dataset, CacheDataset, PersistentDataset
from medlab.registry import DATASETS

DATASETS.register_module(name='Dataset', module=Dataset)
DATASETS.register_module(name='CacheDataset', module=CacheDataset)
DATASETS.register_module(name='PersistentDataset', module=PersistentDataset)



from monai.metrics import DiceMetric, MeanIoU, ConfusionMatrixMetric, HausdorffDistanceMetric, ROCAUCMetric

from medlab.registry import METRICS

METRICS.register_module(name='DiceMetric', module=DiceMetric)
METRICS.register_module(name='MeanIoU', module=MeanIoU)
METRICS.register_module(name='ConfusionMatrixMetric', module=ConfusionMatrixMetric)
METRICS.register_module(name='HausdorffDistanceMetric', module=HausdorffDistanceMetric)
METRICS.register_module(name='ROCAUCMetric', module=ROCAUCMetric)
